/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */

/*
begin_html
See also:
<ul>
<li><a href="matrixutil.h.html">matrixutil.h</a></li>
<li><a href="milutil.h.html">milutil.h</a></li>  
<li><a href="ublas_matrixutil.h.html">ublas_matrixutil.h</a></li>  
</ul>
<b>Up:</b> <a href="index.html">Index</a>
end_html
*/


#ifdef _WIN32
  #include <windows.h>
#else
  #include <stdlib.h>
  #define BOOL int
  #define TRUE 1
  #define FALSE 0
#endif

#include <math.h>
#include <iostream>

#define HAVE_INLINE
#include <gsl/gsl_eigen.h>

#include "matrixutil.h"

using namespace std;
using namespace fq;

//----- Local functions ( optional declaration ) 
double pythag(double a, double b);
void swap(double *a, double *b, int iItems);


// ----------------------------------------------------------
// Swap two doubles - swap(&n,&n) function or SWAP(n,n) macro
double g_nTemp;
#define SWAP(g,h) {g_nTemp=(g);(g)=(h);(h)=g_nTemp;}

#define sign(a,b)   ((b)>=0.0 ? fabs(a) : -fabs(a))
#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))

void swap(double *a, double *b, int iItems){
	double tmp;
	for( int i=0; i<iItems; ++i ) {
		tmp=a[i];
		a[i]=b[i];
		b[i]=tmp;
		}
	}

/*	---------------------------------------------------------------
	Gauss-Jordan linear equation solver
	a[0..n-1][0..n-1] is the input matrix
	b[0..n-1][0..m-1] is input containing the m right-hand side vectors
	on output, a is replaced by its matrix inverse,
	and b is replaced by the corresponding set of solution vectors.
*/

//--  Uses 0 based arrays.
//--  Added comments.  More meaningful variable names.  Better Swap usage.
BOOL gaussj0(double** a, int n, double **b, int m, ostream &msgOut){
	int iIter, iRow, iCol, iMaxCol, iMaxRow;	
	double 	big, dum, pivinv;
	int * indxc = int_alloc1D(n);	// [i] is the column of the i'th pivot element
	int * indxr = int_alloc1D(n);	// [i] is the original row for the i'th pivot 
	int * ipiv = int_alloc1D(n);
	
	//-- Set all pivots to 0 ( false ) 
	for( iRow=0; iRow<n; iRow++) 
		ipiv[iRow]=0;

	iMaxRow = iMaxCol = -1;

	//-- For each iteration, find max and {divide, swap}.
	for (iIter=0; iIter<n; iIter++) {
		//-- For i,j not yet pivoted, find where max abs(a[i][j]) occurs ==> Pivot 
		big=0.0;
		for (iRow=0; iRow<n; iRow++) {
			if (ipiv[iRow] != 1) {
				for (iCol=0; iCol<n; iCol++) {
					if (ipiv[iCol] == 0) { 
						if (fabs(a[iRow][iCol]) > big)
							big=fabs(a[iMaxRow=iRow][iMaxCol=iCol]);
						}
					else if (ipiv[iCol]>1) {
						msgOut << "Fatal err: Singular matrix in gauss-j" << endl;
						free(ipiv);
						free(indxr);
						free(indxc);
						return FALSE;
						}
					}
				}
			}
		if ( iMaxRow == -1 || iMaxCol == -1 ) {
			msgOut << "gauss-j found no pivot" << endl;
			free(ipiv);
			free(indxr);
			free(indxc);
			return FALSE;
			}

		//-- Got the pivot element, interchange rows to force the pivot element to 
		//-- be on the diagonal.  
		++(ipiv[iMaxCol]);
		if (iMaxRow != iMaxCol) {
			swap( a[iMaxRow], a[iMaxCol], n );
			if( b != NULL ) 
				swap( b[iMaxRow], b[iMaxCol], m );
			}

		indxr[iIter]=iMaxRow;
		indxc[iIter]=iMaxCol;
		pivinv = a[iMaxCol][iMaxCol];
		a[iMaxCol][iMaxCol]=1.0;		//-- this is unusual, but is fixed by the next if stmt.
		//-- divide row iCol by big in a and b
		for (iCol=0; iCol<n; iCol++) 
			a[iMaxCol][iCol] /= pivinv;
		for (iCol=0; iCol<m; iCol++) 
			b[iMaxCol][iCol] /= pivinv;

		//-- For each row, subtract a multiple of the maxrow which makes maxcol element 0
		for (iRow=0; iRow<n; iRow++) {
			if (iRow != iMaxCol) {
				dum=a[iRow][iMaxCol];
				a[iRow][iMaxCol]=0.0;
				for (iCol=0; iCol<n; iCol++) 
					a[iRow][iCol] -= a[iMaxCol][iCol]*dum;
				for (iCol=0; iCol<m; iCol++) 
					b[iRow][iCol] -= b[iMaxCol][iCol]*dum;
				}
			}
		}
	
	// unscramble the solution w.r.t. the column interchanges
	// by interchanging columns in the reverse order 
	for( iIter=n-1; iIter>=0; iIter--)
		if (indxr[iIter] != indxc[iIter])
			for (iRow=0; iRow<n; iRow++)
				SWAP( (a[iRow][indxr[iIter]]), (a[iRow][indxc[iIter]]) );

	free(ipiv);
	free(indxr);
	free(indxc);
	return TRUE;
	}


BOOL gaussj_invert(matrix<double>& mat, int n, std::ostream& msgOut){
	return gaussj0(mat, n, NULL, 0, msgOut );
	}

/*	---------------------------------------------------------------
	Pythagoras theorem without destructive underflow or overflow.
	From Numerical Recipes for c 2nd edition 
	Used by svdcmp function.
*/
double pythag(double a, double b){
	a=fabs(a);
	b=fabs(b);
	if( a>b )
		return a*sqrt(1.0+DSQR(b/a));
	if ( b==0.0 )		
		return 0.0;
	return b*sqrt(1.0+DSQR(a/b));
	}

/*	---------------------------------------------------------------
	Singular-value decomposition routine
*/

extern "C" MAXILL_API int __stdcall svdcmp_dll( double **a, int m, int n, double *w, double **v )
{
	return svdecomp( a, m, n, w, v, cerr );
}

int svdecomp(double **a, int m, int n, double *w, double **v, ostream &msgOut){
    int  flag,i,its,j,jj,k,l,nm;
    double  anorm,c,f,g,h,s,scale,x,y,z,*rv1,ww,*aa,*vv;
    const int MAXITER = 30;
    
    rv1 = double_alloc1D(n);
    aa  = double_alloc1D(m);
    vv  = double_alloc1D(n);
    
    g=scale=anorm=0.0;
    for (i=1; i<=n; i++) {
		l=i+1;
		rv1[i-1]=scale*g;
		g=s=scale=0.0;
		if (i <= m) {
			for (k=i; k<=m; k++) scale += fabs(a[k-1][i-1]);
			if (scale) {
				for (k=i; k<=m; k++) {
					a[k-1][i-1] /= scale;
					s += a[k-1][i-1]*a[k-1][i-1];
					}
				f=a[i-1][i-1];
				g = -sign(sqrt(s),f);
				h=f*g-s;
				a[i-1][i-1]=f-g;
				for (j=l; j<=n; j++) {
					for (s=0.0,k=i ;k<=m; k++) 
						s += a[k-1][i-1]*a[k-1][j-1];
					f=s/h;
					for (k=i; k<=m; k++) 
						a[k-1][j-1] += f*a[k-1][i-1];
					}
				for (k=i; k<=m; k++) a[k-1][i-1] *= scale;
				}
			}
		w[i-1]=scale *g;
		g=s=scale=0.0;
		if (i <= m && i != n) {
			for (k=l; k<=n; k++) 
				scale += fabs(a[i-1][k-1]);
			if (scale) {
				for (k=l; k<=n; k++) {
					a[i-1][k-1] /= scale;
					s += a[i-1][k-1]*a[i-1][k-1];
					}
				f=a[i-1][l-1];
				g = -sign(sqrt(s),f);
				h=f*g-s;
				a[i-1][l-1]=f-g;
				for (k=l; k<=n; k++) 
					rv1[k-1]=a[i-1][k-1]/h;
				for (j=l; j<=m; j++) {
					for (s=0.0,k=l; k<=n; k++) 
						s += a[j-1][k-1]*a[i-1][k-1];
					for (k=l; k<=n; k++) 
						a[j-1][k-1] += s*rv1[k-1];
					}
				for (k=l; k<=n; k++) 
					a[i-1][k-1] *= scale;
				}
			}
		anorm=max(anorm,(fabs(w[i-1])+fabs(rv1[i-1])));
		}
    for (i=n; i>=1; i--) {
		if (i < n) {
			if (g) {
				for (j=l; j<=n; j++) 
					v[j-1][i-1]=(a[i-1][j-1]/a[i-1][l-1])/g;
				for (j=l; j<=n; j++) {
					for (s=0.0,k=l; k<=n; k++) 
						s += a[i-1][k-1]*v[k-1][j-1];
					for (k=l; k<=n; k++) 
						v[k-1][j-1] += s*v[k-1][i-1];
					}
				}
			for (j=l; j<=n; j++) 
				v[i-1][j-1]=v[j-1][i-1]=0.0;
			}
		v[i-1][i-1]=1.0;
		g=rv1[i-1];
		l=i;
		}
    for (i=min(m,n); i>=1; i--) {
		l=i+1;
		g=w[i-1];
		for (j=l; j<=n; j++) 
			a[i-1][j-1]=0.0;
		if (g) {
			g=1.0/g;
			for (j=l; j<=n; j++) {
				for (s=0.0,k=l; k<=m; k++) 
					s += a[k-1][i-1]*a[k-1][j-1];
				f=(s/a[i-1][i-1])*g;
				for (k=i; k<=m; k++) 
					a[k-1][j-1] += f*a[k-1][i-1];
				}
			for (j=i; j<=m; j++) 
				a[j-1][i-1] *= g;
			} 
		else {
			for (j=i; j<=m; j++) a[j-1][i-1]=0.0;
			}
		++a[i-1][i-1];
		}
    for (k=n; k>=1; k--) {
		for (its=1; its<=MAXITER; its++) {
			flag=1;
			for (l=k; l>=1; l--) {
				nm=l-1;
				if (fabs(rv1[l-1])+anorm == anorm) {
					flag=0;
					break;
					}
				if (fabs(w[nm-1])+anorm == anorm) 
					break;
				}
			if (flag) {
				c=0.0;
				s=1.0;
				for (i=l; i<=k; i++) {
					f=s*rv1[i-1];
					rv1[i-1]=c*rv1[i-1];
					if (fabs(f)+anorm == anorm) 
						break;
					g=w[i-1];
					h=pythag(f,g);
					w[i-1]=h;
					h=1.0/h;
					c=g*h;
					s = -f*h;
					for (j=1; j<=m; j++) {
						y=a[j-1][nm-1];
						z=a[j-1][i-1];
						a[j-1][nm-1]=y*c+z*s;
						a[j-1][i-1]=z*c-y*s;
						}
					}
				}
			z=w[k-1];
			if (l == k) {
				if (z < 0.0) {
					w[k-1] = -z;
					for (j=1; j<=n; j++) 
						v[j-1][k-1] = -v[j-1][k-1];
					}
				break;
				}
			if (its==MAXITER)
				msgOut << "No convergence in svdecomp0\n";
			x=w[l-1];
			nm=k-1;
			y=w[nm-1];
			g=rv1[nm-1];
			h=rv1[k-1];
			f=((y-z)*(y+z)+(g-h)*(g+h))/(2.0*h*y);
			g=pythag(f,1.0);
			f=((x-z)*(x+z)+h*((y/(f+sign(g,f)))-h))/x;
			c=s=1.0;
			for (j=l; j<=nm; j++) {
				i=j+1;
				g=rv1[i-1];
				y=w[i-1];
				h=s*g;
				g=c*g;
				z=pythag(f,h);
				rv1[j-1]=z;
				c=f/z;
				s=h/z;
				f=x*c+g*s;
				g = g*c-x*s;
				h=y*s;
				y *= c;
				for (jj=1; jj<=n; jj++) {
					x=v[jj-1][j-1];
					z=v[jj-1][i-1];
					v[jj-1][j-1]=x*c+z*s;
					v[jj-1][i-1]=z*c-x*s;
					}
				z=pythag(f,h);
				w[j-1]=z;
				if (z) {
					z=1.0/z;
					c=f*z;
					s=h*z;
					}
				f=c*g+s*y;
				x=c*y-s*g;
				for (jj=1; jj<=m; jj++) {
					y=a[jj-1][j-1];
					z=a[jj-1][i-1];
					a[jj-1][j-1]=y*c+z*s;
					a[jj-1][i-1]=z*c-y*s;
					}
				}
			rv1[l-1]=0.0;
			rv1[k-1]=f;
			w[k-1]=x;
			}
		}
	
 	// Reorders columns of a,v matrices and w vector ordered by abs(w) descending.
	// This appears to be an inefficient swap sort but check more thoroughly.
	// More efficient : Quick sort the W array and a parallel vector of column 
	// destinations then move the a and v columns efficiently.
   for (j=2; j<=n; j++) {
		ww=w[j-1];
		for (i=1; i<=m; i++) 
			aa[i-1]=a[i-1][j-1];
		for (i=1; i<=n; i++) 
			vv[i-1]=v[i-1][j-1];
		k=j-1;
		while ( k > 0 && -fabs(w[k-1])>-fabs(ww) ) {
			w[k+1-1]=w[k-1];
			for (i=1; i<=m; i++) 
				a[i-1][k+1-1]=a[i-1][k-1];
			for (i=1; i<=n; i++) 
				v[i-1][k+1-1]=v[i-1][k-1];
			k--;
			}
		w[k+1-1]=ww;
		for (i=1; i<=m; i++) 
			a[i-1][k+1-1]=aa[i-1];
		for (i=1; i<=n; i++) 
			v[i-1][k+1-1]=vv[i-1];
		}
    
    free(rv1);
    free(aa);
    free(vv);

	return 0;
	}             


// WARNING : a, v are 0 based arrays. wr and wi are 1 based arrays.
BOOL eigen(double** a, int n, double* wr, double* wi, double** v, ostream &msgOut){
  gsl_matrix *Q = gsl_matrix_alloc(n,n);
  for (int i=0; i<n; ++i)
	  for (int j=0; j<n; ++j)
		  gsl_matrix_set(Q,i,j, a[i][j]);
  gsl_matrix_complex *V = gsl_matrix_complex_alloc(n,n);
  gsl_vector_complex *L = gsl_vector_complex_alloc(n);
  gsl_eigen_nonsymmv_workspace *ws = gsl_eigen_nonsymmv_alloc(n);
  gsl_eigen_nonsymmv(Q, L, V, ws);
  for (int i=0; i<n; ++i) {
    wr[i+1] = GSL_REAL(gsl_vector_complex_get(L, i));
    wi[i+1] = 0.0; // GSL_IMAG(gsl_vector_complex_get(L, i));
    for (int j=0; j<n; ++j)
      v[i][j] = GSL_REAL(gsl_matrix_complex_get(V,i,j));
  }
  gsl_matrix_free(Q);
  gsl_matrix_complex_free(V);
  gsl_vector_complex_free(L);
  gsl_eigen_nonsymmv_free(ws);
  return TRUE;
}
